/*
 PQ Kernel 1.0
 
 This algorithm computes the PQ kernel in O(n log n) time based on 
 merge sort. This is a header file to be imported in C/C++ code.
 
 Copyright (C) 2014  Radu Tudor Ionescu, Marius Popescu
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or any
 later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

long merge_sort1(double *a, double *tmp_array, int left, int right)
{
    int i, j, k, center;
    long sl, sr, m;
    
    if (left < right)
    {
        center = (left + right) >> 1;
        sl = merge_sort1(a, tmp_array, left, center);
        sr = merge_sort1(a, tmp_array, center + 1, right);
        
        m = 0;
        i = left;
        j = center + 1;
        while ((i <= center) && (j <= right))
        {
            if (a[j] < a[i])
            {
                m += center + 1 - i;
                j++;
            }
            else
            {
                i++;
            }
        }
        for (i = center + 1; i > left; i--) tmp_array[i - 1] = a[i - 1];
        for (j = center; j < right; j++) tmp_array[right + center - j] = a[j + 1];
        for (k = left; k <= right; k++)
            a[k] = (tmp_array[i] < tmp_array[j]) ? tmp_array[i++] : tmp_array[j--];
        return sl + sr + m;
    }
    else
    {
        return 0;
    }
}
                     
void merge_sort2(double *a, double *b, double *tmp_array_1, double *tmp_array_2, int left, int right)
{
    int i, j, k, center;
    
    if (left < right)
    {
        center = (left + right) >> 1;
        merge_sort2(a, b, tmp_array_1, tmp_array_2, left, center);
        merge_sort2(a, b, tmp_array_1, tmp_array_2, center + 1, right);
        
        for (i = center + 1; i > left; i--)
        {
            tmp_array_1[i - 1] = a[i - 1];
            tmp_array_2[i - 1] = b[i - 1];
        }
        for (j = center; j < right; j++)
        {
            tmp_array_1[right + center - j] = a[j + 1];
            tmp_array_2[right + center - j] = b[j + 1];
        }
        for (k = left; k <= right; k++)
            if ((tmp_array_1[i] < tmp_array_1[j]) || ((tmp_array_1[i] == tmp_array_1[j]) && (tmp_array_2[i] < tmp_array_2[j])))
            {
                a[k] = tmp_array_1[i];
                b[k] = tmp_array_2[i];
                i++;
            }
            else
            {
                a[k] = tmp_array_1[j];
                b[k] = tmp_array_2[j];
                j--;
            }
    }
}

void merge_sort_idx(double *a, int *idx, double *tmp_array_1, int *tmp_array_2, int left, int right)
{
    int i, j, k, center;
    
    if (left < right)
    {
        center = (left + right) >> 1;
        merge_sort_idx(a, idx, tmp_array_1, tmp_array_2, left, center);
        merge_sort_idx(a, idx, tmp_array_1, tmp_array_2, center + 1, right);
        
        for (i = center + 1; i > left; i--)
        {
            tmp_array_1[i - 1] = a[i - 1];
            tmp_array_2[i - 1] = idx[i - 1];
        }
        for (j = center; j < right; j++)
        {
            tmp_array_1[right + center - j] = a[j + 1];
            tmp_array_2[right + center - j] = idx[j + 1];
        }
        for (k = left; k <= right; k++)
            if (tmp_array_1[i] < tmp_array_1[j])
            {
                a[k] = tmp_array_1[i];
                idx[k] = tmp_array_2[i];
                i++;
            }
            else
            {
                a[k] = tmp_array_1[j];
                idx[k] = tmp_array_2[j];
                j--;
            }
    }
}

/* The input X is a matrix of samples (histograms).
   The matrix has n samples of size m, given in a linearized form.
   Samples are stored in contiguous memory zones.
   The function will compute a kernel matrix (with n rows and n columns) 
   between each pair of samples. The returned kernel matrix is also linearized.
 */
double * pqk_matrix(double *X, int m, int n)
{
    double *a, *b, *x, rez, *ker, *tmp_array_1, *tmp_array_2;
    long s, n0, n1, n2, n3, t;
    int i, j, k, l;
    int *idx, *tmp_idx;
    
    x = malloc(m * n * sizeof(double));
    
    idx = malloc(m * sizeof(int));
    tmp_idx = malloc(m * sizeof(int));
    
    a = malloc(m * sizeof(double));
    b = malloc(m * sizeof(double));
    
    tmp_array_1 = malloc(m * sizeof(double));
    tmp_array_2 = malloc(m * sizeof(double));
        
    ker = malloc(n * n * sizeof(double));
    
    if (x == NULL || a == NULL || b == NULL ||
        tmp_array_1 == NULL || tmp_array_2 == NULL ||
        idx == NULL || tmp_idx == NULL || ker == NULL)
    {
        printf("Error: Out of memory.");
        return NULL;
    }
    
    memcpy(x, X, m * n * sizeof(double));

    for (k = 0; k < n; k++)
    {
        memcpy(a, &x[k * m], m * sizeof(double));
        for (i = 0; i < m; i++) idx[i] = i;
        merge_sort_idx(&x[k * m], idx, tmp_array_1, tmp_idx, 0, m - 1);
        
        for (l = k; l < n; l++)
        {
            if (l == k)
            {
                memcpy(b, a, m * sizeof(double));
            }
            else
            {
                memcpy(a, &x[k * m], m * sizeof(double));
                for (i = 0; i < m; i++) b[i] = x[l * m + idx[i]];
            }
            
            merge_sort2(a, b, tmp_array_1, tmp_array_2, 0, m - 1);
            
            // Compute the total number of pairs
            n0 = (m * (m - 1)) >> 1;
            
            // Compute the number of equal pairs in a
            i = 0;
            n1 = 0;
            while (i < m - 1)
            {
                t = 1;
                j = i + 1;
                while (a[j] == a[i]) {j++; t++;}
                n1 += (t * (t - 1)) >> 1;
                i = j;
            }
            
            // Compute the number of equal pairs in a and b
            i = 0;
            n3 = 0;
            while (i < m - 1)
            {
                t = 1;
                j = i + 1;
                while ((a[j] == a[i]) && (b[j] == b[i])) {j++; t++;}
                n3 += (t * (t - 1)) >> 1;
                i = j;
            }
            
            // Compute the number of discordant pairs
            s = merge_sort1(b, tmp_array_2, 0, m - 1);
            
            // Compute the number of equal pairs in b
            i = 0;
            n2 = 0;
            while (i < m - 1)
            {
                t = 1;
                j = i + 1;
                while (b[j] == b[i]) {j++; t++;}
                n2 += (t * (t - 1)) >> 1;
                i = j;
            }
            
            rez = (double) (n0 + n3 - n1 - n2 - (2 * s));
            ker[l * n + k] = rez;
            ker[k * n + l] = ker[l * n + k];
        }
    }
    
    free(x);
    free(a);
    free(b);
    free(idx);
    free(tmp_idx);
    free(tmp_array_1);
    free(tmp_array_2);
    
    return ker;
}

/* The two input parameters A and B are two samples (or histograms).
   Function will return the PQ kernel between the two samples.
 */
double pqk(double *A, double *B, int n)
{
    double *a, *b, rez, *tmp_array_1, *tmp_array_2;
    long s, n0, n1, n2, n3, t;
    int i, j, k, l;
    
    a = malloc(n * sizeof(double));
    b = malloc(n * sizeof(double));
    
    tmp_array_1 = malloc(n * sizeof(double));
    tmp_array_2 = malloc(n * sizeof(double));
    
    if (a == NULL || b == NULL ||
        tmp_array_1 == NULL || tmp_array_2 == NULL)
    {
        printf("Error: Out of memory.");
        return 0.0;
    }
    
    memcpy(a, A, n * sizeof(double));
    memcpy(b, B, n * sizeof(double));
    merge_sort2(a, b, tmp_array_1, tmp_array_2, 0, n - 1);
    
    // Compute the total number of pairs
    n0 = (n * (n - 1)) >> 1;
    
    // Compute the number of equal pairs in a
    i = 0;
    n1 = 0;
    while (i < n - 1)
    {
        t = 1;
        j = i + 1;
        while (a[j] == a[i]) {j++; t++;}
        n1 += (t * (t - 1)) >> 1;
        i = j;
    }
    
    // Compute the number of equal pairs in a and b
    i = 0;
    n3 = 0;
    while (i < n - 1)
    {
        t = 1;
        j = i + 1;
        while ((a[j] == a[i]) && (b[j] == b[i])) {j++; t++;}
        n3 += (t * (t - 1)) >> 1;
        i = j;
    }
    
    // Compute the number of discordant pairs
    s = merge_sort1(b, tmp_array_2, 0, n - 1);
    
    // Compute the number of equal pairs in b
    i = 0;
    n2 = 0;
    while (i < n - 1)
    {
        t = 1;
        j = i + 1;
        while (b[j] == b[i]) {j++; t++;}
        n2 += (t * (t - 1)) >> 1;
        i = j;
    }

    rez = (double) (n0 + n3 - n1 - n2 - (2 * s));
    
    free(a);
    free(b);
    free(tmp_array_1);
    free(tmp_array_2);
    
    return rez;
}

/* The inputs X and Y are two matrices of samples (histograms).
   The matrix X has n samples of size m, and Y has nx samples of size m.
   In both matrices, samples are stored in contiguous memory zones.
   Both matrices are given in a linearized form.
   The function will compute a kernel matrix (with nx rows and n columns)
   between pairs of samples a and b, where b is taken from X and a from Y.
   The returned kernel matrix is also linearized.
 */
double * pqk_cross_matrix(double *X, double *Y, int m, int n, int nx)
{
    double *tmp1, *tmp2;
    double *a, *b, *x, *xx, rez, *ker, *tmp_array_1, *tmp_array_2;
    long s, n0, n1, n2, n3, t;
    int i, j, k, l;
    int *idx, *tmp_idx;
    
    x = malloc(m * n * sizeof(double));
    xx = malloc(m * nx * sizeof(double));

    idx = malloc(m * sizeof(int));
    tmp_idx = malloc(m * sizeof(int));
    
    a = malloc(m * sizeof(double));
    b = malloc(m * sizeof(double));
    
    tmp_array_1 = malloc(m * sizeof(double));
    tmp_array_2 = malloc(m * sizeof(double));
    
    ker = malloc(nx * n * sizeof(double));
    
    if (x == NULL || xx == NULL ||
        a == NULL || b == NULL ||
        tmp_array_1 == NULL || tmp_array_2 == NULL ||
        idx == NULL || tmp_idx == NULL || ker == NULL)
    {
        printf("Error: Out of memory.");
        return NULL;
    }
    
    memcpy(x, X, m * n * sizeof(double));
    memcpy(xx, Y, m * nx * sizeof(double));
    
    for (k = 0; k < nx; k++)
    {
        for (i = 0; i < m; i++) idx[i] = i;
        merge_sort_idx(&xx[k * m], idx, tmp_array_1, tmp_idx, 0, m - 1);
        
        for (l = 0; l < n; l++)
        {
            memcpy(a, &xx[k * m], m * sizeof(double));
            for (i = 0; i < m; i++) b[i] = x[l * m + idx[i]];
            
            merge_sort2(a, b, tmp_array_1, tmp_array_2, 0, m - 1);
            
            // Compute the total number of pairs
            n0 = (m * (m - 1)) >> 1;
            
            // Compute the number of equal pairs in a
            i = 0;
            n1 = 0;
            while (i < m - 1)
            {
                t = 1;
                j = i + 1;
                while (a[j] == a[i]) {j++; t++;}
                n1 += (t * (t - 1)) >> 1;
                i = j;
            }
            
            // Compute the number of equal pairs in a and b
            i = 0;
            n3 = 0;
            while (i < m - 1)
            {
                t = 1;
                j = i + 1;
                while ((a[j] == a[i]) && (b[j] == b[i])) {j++; t++;}
                n3 += (t * (t - 1)) >> 1;
                i = j;
            }
            
            // Compute the number of discordant pairs
            s = merge_sort1(b, tmp_array_2, 0, m - 1);
            
            // Compute the number of equal pairs in b
            i = 0;
            n2 = 0;
            while (i < m - 1)
            {
                t = 1;
                j = i + 1;
                while (b[j] == b[i]) {j++; t++;}
                n2 += (t * (t - 1)) >> 1;
                i = j;
            }
            
            rez = (double) (n0 + n3 - n1 - n2 - (2 * s));
            ker[l * nx + k] = rez;
        }
    }
    
    free(x);
    free(xx);
    free(a);
    free(b);
    free(idx);
    free(tmp_idx);
    free(tmp_array_1);
    free(tmp_array_2);

    return ker;
}
